""" Constants file for Thermo Fisher Qubit Flex Parser"""

# Instrument constants
SOFTWARE_NAME = "Qubit Flex software"
MODEL_NUMBER = "Qubit Flex"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"
BRAND_NAME = "Qubit"
DEVICE_TYPE = "fluorescence detector"
DISPLAY_NAME = "Thermo Fisher Scientific Qubit Flex"

# Error constants
VALUE_ERROR = "Unable to find value for column"
