DISPLAY_NAME = "Thermo Fisher Scientific VISIONlite"
DEVICE_TYPE = "spectrophotometer"
DETECTION_TYPE = "absorbance"
SOFTWARE_NAME = "VISIONlite"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"

# Errors
UNSUPPORTED_KINETIC_MEASUREMENTS_ERROR = (
    "Kinetic measurements are not supported at this time."
)

HEADER_COLS = [
    "File Name",
    "Sample Name",
    "Analyst",
    "Measurement Time",
]
SAMPLE_NAME_COLS = ("Sample Name", "Sample name", "sample name", "Well Position")
