""" Constants file for ThermoFisher Genesys 30 Adapter"""

# Instrument software details
GENESYS_SOFTWARE = "Genesys 30 instrument software"
DISPLAY_NAME = "Thermo Fisher Scientific Genesys30"
MODEL_NUMBER = "Genesys 30"
BRAND_NAME = "Genesys"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"
DEVICE_TYPE = "absorbance detector"

# Error messages:
MEASUREMENT_TIME_MISSING = "Unable to find measurement time from instrument file"
