DEVICE_IDENTIFIER = "Nanodrop"
DEVICE_TYPE = "absorbance detector"
MODEL_NUBMER = "Nanodrop Eight"

# These may be reported in the results, and are stored as calculated data
ABSORBANCE_RATIOS = [
    (260, 280),
    (260, 230),
]

CONCENTRATION_UNITS = ["ng/μL", "μg/mL", "μg/µL", "mg/mL", "μg,mL"]
